from flask import Flask, jsonify, abort, request, jsonify
from pyproteinsext import uniprot as pExt
from . import collectionProxy as redisCollection
import re


UNIPROT_COLLECTION=None
# temp hack
REDIS=False

def cleanup(rh=None, rp=None):
    redisCollection.bootstrap(host=rh, port=rp)
    redisCollection.cleanup()

def startup(xmlUniprot, redis=False, rh=None, rp=None):

    global UNIPROT_COLLECTION, REDIS
   # UNIPROT_COLLECTION = pExt.EntrySet(collectionXML=xmlUniprot)
   
    if xmlUniprot:
        print(f"Loading XML ressource {xmlUniprot} ...")
        _ = pExt.EntrySet(collectionXML=xmlUniprot)
        print("before", len(_))
        _.add('P27573')
        print('after', len(_))
        
        UNIPROT_COLLECTION = _    
    if redis:
        REDIS=True
        redisCollection.bootstrap(host=rh, port=rp)        
        if xmlUniprot:
            redisCollection.convert(_)
        UNIPROT_COLLECTION = redisCollection
        
    print("Uniprot storage service listening")
 
    app = Flask(__name__)

    app.add_url_rule('/model', 'model', model)
    
    app.add_url_rule( '/uniprot/<uniprotID>', 'getProtein', getProtein,
                      methods = ['GET'] ) 
    
    app.add_url_rule( '/uniprots', 'getProteins', getProteins,
                      methods = ['POST'] )

    app.add_url_rule( '/length', 'length', length,
                      methods = ['GET'] ) 
    
    app.add_url_rule( '/uniprot/list', 'list', listProtein)
    app.add_url_rule( '/uniprot/list/<interval>', 'listProtein', listProtein)
    app.add_url_rule('/missing', 'getMissing', get_missing, methods = ['POST'])
    #app.add_url_rule('/unigo/<taxid>', 'view_unigo', view_unigo)
    
    return app

def listProtein(interval=':1000'):
    def parseInterval(string): 
        """Parse a slice-like expression"""       
        m1 = re.match("^([\d]+):{0,1}$", string)
        if m1:
            return (0, int(m1[1]))
        m2 = re.match("^:([\d]+)$", string)
        if m2:
            return (int(m2[1]), None)
        m3 = re.match("^([\d]+):([\d]+)$", string)
        if m3:
            return (int(m3[1]), int(m3[2]))
    
    cstart, cstop = parseInterval(interval)

    listIDs = redisCollection.getSliceIDs(cstart, cstop)
    
    return jsonify( {"entryIDs" : listIDs} )



def length():
    global UNIPROT_COLLECTION
    if REDIS:
        length = UNIPROT_COLLECTION.length()
    else:
        length = len(UNIPROT_COLLECTION)
    print(f"Current Collection size ${length}")
    return jsonify( {"totalEntry" : length} )

def getProtein(uniprotID):
    print(f"Seeking {uniprotID}")
    oProtein = UNIPROT_COLLECTION.get(uniprotID)
    if not oProtein:
        abort(404)
    return jsonify( { uniprotID : oProtein.toJSON() } )

def _load_uniprot_ids_request():
    if not request.is_json:
        print("Post data not json")
        abort(422)
    data = request.get_json()
    if not 'uniprotIDs' in data:
        print("Post data lacks uniprotIDs key")
        abort(422)
    return data


def getProteins():
    data = _load_uniprot_ids_request()
    ## implements redis
    results = {}
    validCnt = 0
    # tmp hack
    if not REDIS:
        for id in data['uniprotIDs']:
            _ = UNIPROT_COLLECTION.get(id)
            results[id] = _.toJSON() if not _ is None else None
            validCnt = validCnt + 1 if results[id] else validCnt
    else:
        print("Fetching many via redis")
        print(data['uniprotIDs'][:10])
        #test = [e.id if e else None for e in UNIPROT_COLLECTION.mget(data['uniprotIDs'], raw=False)]
        #print(test[:10])
        #indices = [i for i, x in enumerate(test) if x == None]
        #print("Missing", indices)

        for uniprot_id, e in zip( data['uniprotIDs'],\
                                  UNIPROT_COLLECTION.mget(data['uniprotIDs'], raw=False) 
                                ):
            results[ uniprot_id if not e else e.id ] = e.toJSON() if not e is None else None
            validCnt = validCnt + 1 if e is not None else validCnt

    print(f"Returning { validCnt } valid elements")
    return jsonify(results)

def get_missing():
    '''Return missing proteins'''
    print("Check availability")
    data = _load_uniprot_ids_request()
    redis_proteins = [e.id if e else None for e in UNIPROT_COLLECTION.mget(data['uniprotIDs'], raw=False)]
    if None in redis_proteins:
        missing_indices = [i for i, x in enumerate(redis_proteins) if x == None]
        missing_proteins = [data['uniprotIDs'][i] for i in missing_indices]
        return jsonify(missing_proteins)
    return jsonify([])

def model():
    #pyproteinsext.model()
    return "Hello world"


import pyproteinsext.proteinContainer
import gzip


def parse(inputFile=None):
    """Will parse fasta file to create Container
    
    :param inputFile: path of fasta file to parse. 
    :return: fastaContainerFactory.Container
    """
    mainContainer = Container()
    fnOpen = open
    t = 'r'
    if inputFile.endswith('gz'):
        fnOpen = gzip.open
        t = 'rt'
    try:
        f = fnOpen(inputFile, t)
    except IOError:
        print("Could not read file:", inputFile)
        return Container()

    instance = []
    with f:
        for l in f:
            if l.startswith(">"):
                if instance:
                    mainContainer = mainContainer.addParsing(Container(input=instance))
                instance = []
            instance.append(l.rstrip())
        mainContainer = mainContainer.addParsing(Container(input=instance))
    return mainContainer


def _parseBuffer(input):
    header = input[0].lstrip(">")
    prot = header.split(" ")[0]
    seq = "".join(input[1:])
    obj = Fasta(prot, header, seq)
    return {prot: obj}


class Container(pyproteinsext.proteinContainer.Container):
    """ Iterable container, that stores all Fasta objects """
    def __init__(self, input=None):
        super().__init__(_parseBuffer, input)


class Fasta():
    """
    Fasta object stores informations about one fasta entry. 

    :param prot: Protein id, first element of header, before first space
    :type prot: str
    :param seq: Amino acids or nucleotides sequence
    :type seq: str
    :param header: Fasta header
    :type header: str
    """

    def __init__(self, prot, header, seq):
        """ Constructor method """
        self.prot = prot
        self.seq = seq
        self.header = header

    def get_subsequence(self, start, end):
        """
        Get subsequence of entry, from start to end position
        
        :param start: Start position
        :type start: int
        :param end: End position
        :type end: int
        :return: subsequence
        :rtype: str
        """
        return self.seq[start-1:end]



from .redisClient import listUniprotKey, getUniProtEntry,\
    removeEntries, setDatabaseParameters, storeEntry,\
    storeManyEntries, mgetUniProtEntry
from progressbar import Percentage, Bar, ETA, AdaptiveETA, ProgressBar, UnknownLength, Counter, Timer
#from progressbar import ProgressBar, Bar, Counter, Timer, ETA, Percentage, RotatingMarker
def bootstrap(**kwargs):
    print("Bootstraping uniprot Entry redis Collection")
    setDatabaseParameters(**kwargs)
    print(f"{__len__()} uniprot entries were found in store")

def cleanup(displayCnt=True): 
    cnt = 0
    _ = []

    widgets = ['Deleting ', Counter(), ' ', Percentage(),\
               ' ', Bar(),\
               ' ', Timer(),\
               ' ', AdaptiveETA()]
    with ProgressBar(widgets=widgets, maxval= __len__()\
        if displayCnt else UnknownLength\
        , redirect_stdout=True) as bar:
        for key in listUniprotKey():
            cnt += 1
            _.append(key)
            if cnt % 50 == 0:
                try:
                    remove(_)
                except KeyError:
                    print("Missing keys to delete")
                _ = [] 
                bar.update(cnt)
        if _:
            try:
                remove(_)
            except KeyError:
                print("Missing keys to delete")                
            bar.update(cnt)

def convert(entryIterator, bulkSize=250):
    _ = []
    cnt=0
    #print(">>", entryIterator, "<<")
    widgets = ['Loading ',  Counter(), ' ', Percentage(),\
               ' ', Bar(),\
               ' ', Timer(),\
               ' ', AdaptiveETA()]
    with ProgressBar(widgets=widgets,maxval=len(entryIterator)\
        , redirect_stdout=True) as bar:
        for e in entryIterator:
            _.append(e)
            cnt+=1
            if cnt % bulkSize == 0:
                storeManyEntries(_)                
                bar.update(cnt)
                _ = []
        if _:
            storeManyEntries(_)                
            bar.update(cnt)

def getSliceIDs(cstart=0, cstop=None):
    cnt = 0
    results = [ ]
    for e in __iter__():
        if not cstop is None:
            if cnt == cstop:
                return results
        
        if cnt >= cstart:           
            results.append(e.id)
        cnt += 1
    
    return results

def add(e):
    """Add a single uniprot entry to the store"""
    try :
        storeEntry(e)
    except KeyError as err :
        print(f"No need to add {e.id} already in store")

# Proof of concept for deletion
# Improvment stage 1 : send slice to removeEtnries. current delete decorator is one by one
# Imorvement stage 2 : use pipeline or scriptiong in redisCLient
def remove(uniprotIDs=None):
    print(f"Removing {uniprotIDs}")
    
    if uniprotIDs: # slow delete all 
        if type(uniprotIDs) == list:        
            removeEntries(uniprotIDs)
        elif type(uniprotIDs) == str:
            removeEntries([uniprotIDs])
    
# Get Object
# deserialize
def get(uniprotID, raw=False):
    _ =  getUniProtEntry(uniprotID)
    #print(f"PWEEPWEE {_}")
    return _

def mget(uniprotIDs, raw=False):
    _ =  mgetUniProtEntry(uniprotIDs, raw=raw)
    return _

# Not sure its possible at module level
def __iter__():
    chunckSize = 250
    _ = []
    cnt = 0
    for _id in listUniprotKey():
        cnt += 1 
        _.append(_id)
        if cnt % chunckSize == 0:            
            for e in mget(_):
                yield e
            _ = []
    if _:
        for e in mget(_):
            yield e 
    print(f"Completed total {cnt} iterations")

# Not possible at module level
# TypeError: object of type 'module' has no len()
def __len__():
    cnt = 0
    for _id in listUniprotKey(): 
        cnt += 1
    return cnt

def length():
    cnt = 0
    for _id in listUniprotKey(): 
        cnt += 1
    return cnt





# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyproteinsext', 'pyproteinsext.database', 'pyproteinsext.services.uniprot']

package_data = \
{'': ['*'],
 'pyproteinsext': ['notebooks/*',
                   'notebooks/.ipynb_checkpoints/*',
                   'notebooks/data/*']}

install_requires = \
['biopython>=1.79,<2.0', 'bs4>=0.0.1,<0.0.2', 'pyproteins>=2.0.0,<3.0.0']

setup_kwargs = {
    'name': 'pyproteinsext',
    'version': '3.1.2',
    'description': "'Extending pyproteins for bioinformatics tools&services'",
    'long_description': None,
    'author': 'glaunay',
    'author_email': 'pitooon@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
    Build a database of fasta entries using file system directory architecture
    Usage:
        uniprotFastaFS.py <location> [--cluster <bigFile> --radius <radius>]
        uniprotFastaFS.py <location> [--nodes <volNumExp> --threads <nWorker> --size <size>]
        uniprotFastaFS.py <location> [--input <file> --size <size> --auto]
        uniprotFastaFS.py <location> [--get <proteinID> --out <fileOut>]
        uniprotFastaFS.py <location> (-i | --info)
        uniprotFastaFS.py --view <file>

    Options:
        -h --help                             Show this screen
        -d <file>, --input <file>             MultiFasta input file
        -s <size>, --size <size>              Volume size [default: 500]
        -a, --auto                            Build without asking confirmation
        -i, --info                            Display Database statistics
        -v <file>, --view <file>              Display Input file statistics
        -g <proteinID>, --get <proteinID>     Identifier to retrieve data
        -o <fileOut>, --out <fileOut>         output file [default: stdout]
        -c <file>, --cluster <file>           archive file to dispatch to nodes
        -r <radius>, --radius <radius>        maximum number or entries per nodes
        -n <volNumExp>, --nodes <volNumExp>   Building the volumes of marching regExp  number nodes
        -t <nWorker>, --threads <nWorker>     Set the number of workers for parrallel building of node volumes

"""
#
# Wrapper to a file system base database of fasta sequences
# NB: insert shoukld 1st stat nodes to modify smallest one
import re
import os
import glob
import gzip
import sys
import shutil
from multiprocessing import Pool, TimeoutError
#import zipfile


from docopt import docopt

# iterate over entries found in file
# forEach insertID 
# then
# forEach load
def batchBuild(rootFolder, fileName, Nsize=500, alreadyIndexed=False):

    if not alreadyIndexed:
        scratchInsert(rootFolder, fileName, Nsize)

    scratchLoad(rootFolder, fileName)

    print("Scratch indexing and data dumping complete, archiving (Vsize=" + str(Nsize) + ")")
    a=0
    for iFile in glob.iglob(rootFolder + '/**/data.raw', recursive=True):
        with open(iFile, 'r') as fInput:
            a += 1
            jFile = iFile.replace('.raw','.gz')
            with gzip.open(jFile, 'wb') as fOutput:
                for l in fInput:    
                    fOutput.write(l.encode('utf-8'))
    
        os.remove(iFile)
    print("Archiving of " + str(a) + " volumes successfull")

def preview(fileName):
    c = 0
    for elem in fileCrawl(fileName):
        c += 1
    print("Input File features " + str(c) + " entries ")


def scratchInsert(rootFolder, fileName, Nsize):
    cnt = 0
    
    for elem in fileCrawl(fileName):
        _insertID(elem['id'], rootFolder, Nsize, BuildStep=True)
        cnt += 1

    print(str(cnt) + " elements inserted from scratch")

def scratchLoad(rootFolder, fileName):
    print("Adding content from scratch")
    for elem in fileCrawl(fileName):
        _load(rootFolder, elem['id'], elem['content'], BuildStep=True)

def stat(rootFolder):
    stat = []
    print("Computing stats w/ file pattern " + rootFolder + '/**/index.txt')
    t=0
    for iFile in glob.iglob(rootFolder + '/**/index.txt', recursive=True):
        #print (iFile)
        c = 0
        with open(iFile, 'r') as f:
            for l in f:
                c += 1
        stat.append((iFile[len(rootFolder)+1:].rstrip('/index.txt').replace('/',''), c))
        t += c
    print("Total indexed entries " + str(t))
    return sorted(stat, key=lambda x:x[1], reverse=True) 

## Open flat file eventually zip compress and pass it to fastaStrea Stream yielder
def fileCrawl(fileName):
    fOpen = open
    mode = 'r'
    _args = {}
    if fileName.endswith('gz'):
        fOpen = gzip.open
        mode = 'rt'
        _args = { "encoding" : 'utf-8' }
       #print("Reading from zipped source")
        
    with fOpen(fileName, mode, **_args) as f:
        fs = fastaStream(f)
        for d in fs:
            yield d

# Stream Read and iterate other each fasta record
class fastaStream(object):
    def __init__(self, iStream):
        self.iStream = iStream

    def __iter__(self):
        datum = {'id' : None, 'content' : None}
        for l in self.iStream:
            if l.startswith(">"):
                if datum['id']:
                    yield datum
            
                datum['id'] = l.split('|')[1]
                datum['content'] = l
            else:
                datum['content'] += l
        yield datum



# Given a collection of fasta and a splitLvl
# Create the required directories 
# Dump the corresponding gziped fasta file and index txt file
# WE LL SEE

def _load(workDir, _id, content, BuildStep=False):
    vDir, lvl, vName, gotcha = _getElemDir(_id, workDir)
    if not vDir:
        raise ValueError("No suitable volume for " + _id + " content")
    
    vInfo = _dirIndex(vDir, vName)
    elemInfo = [ (e[0], e[2]) for e in vInfo if e[0] == _id ] # Filter out vInfo of elemnt to load
    if len(elemInfo) > 1:
        raise TypeError("Volume key duplicate")

    if not elemInfo:
        raise ValueError("Matching volume does not hold key  " + _id)
    elemInfo = elemInfo[0]
    
    if elemInfo[1] == '1' and not BuildStep:
        print (elemInfo[0] + ' data already loaded' )
        return True

    updateVolumeDatum(vDir, _id, content, BuildStep=BuildStep)
    return True


def updateVolumeDatum(vDir, _id, content, BuildStep=False):
    dataFile = vDir + '/data.gz' if not BuildStep else vDir + '/data.raw'
# Build call -> aappend to flat file    
    if BuildStep:
        with open(dataFile, 'a') as f:
            f.write(content)
        return True

# Updating archived data
    toDump = content
    if os.path.isfile(dataFile):
        with gzip.open(dataFile, 'rt', encoding='utf-8') as fRead :
            toDump = ('').join([ l for l in fRead ])
        toDump += content    

    with gzip.open(dataFile, 'wb') as f:
            f.write(toDump.encode('utf-8'))
    
    # Updating index file
    rawInfo = []
    with open(vDir + '/index.txt', 'r') as f:
        for l in f:
            _buf = l.split()
            if _buf[0] ==  _id:
                _buf[1] = '1'
            rawInfo.append(_buf)
   
    with open(vDir + '/index.txt', 'w') as f:
       f.write( '\n'.join([ '\t'.join(e) for e in rawInfo ])  + '\n' )
    
    return True

# Create desired new directory w/ index file of elem provided list
def createVolume(volumeID, rootDir, elem=[], data=[]):
    newVolumePath = rootDir + '/' + ("/").join(volumeID)
    
    os.mkdir(newVolumePath)
    with open(newVolumePath + '/index.txt' , 'w') as f:
        for e in elem:
            f.write('\t'.join(e) + '\n')

    if data:
        with gzip.open(newVolumePath + '/data.gz', 'wb') as f:
            f.write('\n'.join(data).encode('utf-8'))
    
    return newVolumePath

## If queryKeys is provided returns single key dictionary
def _dataDict(dirPath, queryKey=None):
    if not os.path.isfile(dirPath + '/data.gz'):
        return None
    _d = {}
    
    for d in fileCrawl(dirPath + '/data.gz'):
        if queryKey:
            if d['id'] == queryKey:
                return { queryKey : d['content'] }
            continue
        _d[d['id']] = d['content']
    
    return _d

def _insertID(elem, rootDir, N=2, BuildStep=False):
   # if exists(elem, rootDir) and not Force:
   #     print (elem + ' already in database')
   #     return True
    
    
    ## Look for deepest possible
    dirPath, dirLvl, vName, gotcha = _getElemDir(elem, rootDir)
    ## Create it if none suitable
    if not dirPath:
        dirPath = createVolume(elem[0], rootDir) # Special case back to root
    ## Check the size of the volume were about to insert into
    #vInfo = _dirIndex(dirPath, vName)
    #if elem in [ e[0] for e in vInfo ]:
    if gotcha:
        print(elem + ' already in DB no ID to insert')
        return True

    vInfo = _dirIndex(dirPath, vName)
    #print ("About to insert id in a " + str(len(dirElem)) + ' elements index')
    ## If size limits distribute the current volume content into subvolumes
    if len(vInfo) >= N:
        #print("Volume "+ dirPath + ' reached size limit (' + str(N) + ') spliting it' )
        # LOADING Fasta Record
        heavyData = _dataDict(dirPath)
        if shotgun(rootDir, vInfo, dirLvl, N, vName, heavyData):
    ## Redo the insertion attempt
            #print("Shotgun completed attempting inserting again")
            _insertID(elem, rootDir, N)
            return True
        else :# Is elem to add in its max suffix volume
            if len(elem) > len(vName) + 1: # Too long to be indexed here
            #    print(str(vName) + " No spread out of shotgun => creating volume " + elem[:len(vName) + 1] + " to index " + elem)
                createVolume(elem[:len(vName) + 1], rootDir, elem=[elem])
                return True
            #print("No spread of " + vName +  " possible appending terminal suffix element " + elem)
#else: # We were not able to shotgun current index, bc all indexed elements have max len suffix
# We inject           
#        return True
    ## Actual insertion -> index update
    with open(dirPath + '/index.txt', 'a') as f:
        status = 0 if not BuildStep else 1
        f.write(elem + '\t' + str(status) + '\n')
    #print(elem + ' appened to ' + dirPath + '/index.txt')
    return True
# Separate element within current directory 
# which are already stored in their longest possible suffix => setA
# From the other => setB
# setB will be split in subdirectories
# setA will overwrite index file of current directory 
def shotgun(rootDir, elemList, dirLvl, N, cVolName, hRecord):
    #print( str(dirLvl) + ' ' + cVolName + " , LOADING " + str(elemList))
    small = []
    long = []
    for e in elemList:
        t = small if e[1] else long
        t.append( (e[0], e[2]) ) ## Storin name and load status

    #print('small ' + str(small) + ' // long' + str(long))
    
    if len(long) == 0: # Cnat do anything
        return False
    
    #print('Spreading ' + str(len(long)) + ' non max suffix element' )
    #print('XX ' + str(dirLvl))
    
    
    newLots = {}
    for e in long:
        k = cVolName + e[0][dirLvl]
        if not k in newLots:
            newLots[k] = []
        newLots[k].append(e)


    for k in newLots:
        newData = []
        if hRecord:
            newData = [ hRecord[ e[0] ]for e in newLots[k] if e[0] in hRecord ]
        createVolume(k, rootDir, elem=newLots[k], data=newData)
    # Erase previous index with max suffix name only
    upIndex = rootDir + '/' + ('/').join(cVolName) + '/index.txt'
    upData = rootDir + '/' + ('/').join(cVolName) + '/data.gz'
    if small:
        with open(upIndex, 'w') as f:
            for e in small:
                f.write('\t'.join(e) + '\n')
        
        if hRecord:
            newDataSmall = [ hRecord[ e[0] ] for e in small if e[0] in hRecord ]
            with gzip.open(upData, 'wb') as f:
                f.write('\n'.join(newDataSmall).encode('utf-8'))

    else:
        os.remove(upIndex)
        if os.path.isfile(upData):
            os.remove(upData)
    return True
        
    #print("Should add " + str(newLetters) + ' to ' + rootDir)
    

def get(elem, rootDir):
    
    dirPath, dirLvl, vName, gotcha = _getElemDir(elem, rootDir)
    if not dirPath:
        raise ValueError("No suitable volume for " + elem + " content")
    
    if not gotcha:
        return None

    d = _dataDict(dirPath)
    return d[elem]


def exists(elem, rootDir):
    dirPath, dirLvl, vName, gotcha = _getElemDir(elem, rootDir)
    return gotcha
 #  if not dirPath:
 #       return False
 #   return elem in [ e[0] for e in _dirIndex(dirPath, vName) ]

# "(elem, isLeave)"
# Eg : A/B/index.txt
# ABC is a leave, no deeper directory suffixin possible
# ABFE is not

def _dirIndex(cDir, vName):
    iFile = cDir + '/index.txt'
    
    if not os.path.isfile(iFile):
        #print('No index file at ' + cDir)
        return []
        #raise ValueError('No index file at ' + cDir)
    data = []
    with open(iFile, 'r') as f:
        for l in f:
            [eName, eLoad] = l.split()
            elemSuf = eName[:-1] # Getting rid of carraige return and last char --> longest possible suffix
            isLeave = str(elemSuf) == str(vName)
            data.append( (eName, isLeave, eLoad) )
    
    return data
    
# Provided a fastaElem,
# Pool version, 
# returns
# - the directory where it should be located (deepest possible)
# - the subtree level
# - the list of entries in this volume
def _getElemDir(_id, rootDir):
    #print("Looking for a volume for request " + _id)
    defVal = [None, 0, None, False]

    nodePath = glob.iglob(rootDir + '/node_*/') if glob.glob(rootDir + '/node_*/') else [ rootDir + '/' ]    
    hitPaths = [ _getElemDirNode(iNode, _id) for iNode in nodePath  ]
   
    
    for iDir in sorted( hitPaths, key=lambda x : x[1] ) : 
        if not iDir[0]: # Current node do not lay a single lvl suitable directory
            continue
        if iDir[1] > defVal[1]: ### No check on max size of current Node
            defVal = list(iDir) + [False]

        #print (iDir + " :: content")
        if _id in [ e[0] for e in _dirIndex(iDir[0], iDir[2]) ]:
            ans = list(iDir) + [True]
            return ans
   # [ for iDir in hitPaths ]
    #print ( sorted( hitPaths, key=lambda x : x[1] ) )
    return defVal

## Given one node folder tree get deepest matching directory path
def _getElemDirNode(rootDir, _id):
    offset = -1
    while len(_id) + offset > 0:
        cPath = rootDir + ("/").join(_id[:offset])
        #print(cPath)
        if os.path.isdir(cPath):
            lvl = len(_id) + offset
            #print("Got following volume location : " + cPath + ' at lvl ' + str(lvl))
            return (cPath, lvl, str(_id[:offset]))
        offset-=1
    
    return (None, 0, None)

# Will split an input file into subfile of maxSize content, and zip them
def setNodes(rootFolder, inputFile, maxSize):
    nodeCounter = 0
    fCurrent = None
    fCurrName = None

    def _wrap(c, fName, fOut):
        fOut.close()
        with open(fName, 'rb') as f_in:
            with gzip.open(rootFolder + "/node_" + str(c) + ".gz", 'wb') as f_out:
                shutil.copyfileobj(f_in, f_out)

        #gzip.GzipFile(rootFolder + "/node_" + str(c) + ".gz", mode='w').write(fName)
        os.remove(fName)

    for eNum, elem in enumerate( fileCrawl(inputFile) ):
        if eNum % int(maxSize) == 0:
            
            if fCurrent: # close and zip the current one, then remove the flat version
               _wrap(nodeCounter, fCurrName, fCurrent)
            # Open the new current one
            nodeCounter += 1
            fCurrName = rootFolder + "/node_" + str(nodeCounter) + ".raw"
            fCurrent = open(fCurrName, "w" )
        fCurrent.write(str( elem['content']) )

    if fCurrent:
        _wrap(nodeCounter, fCurrName, fCurrent)
       
    print ("needs " + str(nodeCounter) + " volumes for maxSize of " + str(maxSize))
    

def pClusterBuild(inputPack):
    fileSource = inputPack[0]
    kwargs = inputPack[1]

    rootFolder = os.path.dirname(fileSource)
    nodeID = re.findall(r'[\d]+',os.path.basename(fileSource))[0]
    nodeRootFolder = rootFolder + '/node_' + str(nodeID)
    os.makedirs(nodeRootFolder)
    batchBuild(nodeRootFolder, fileSource, **kwargs)


if __name__ == "__main__":
    arguments = docopt(__doc__, version='uniprotFastaFSDB 2.0')
    #print(arguments)
    if arguments['--input']:
        if arguments['--auto']:
            print ("Automatic Building confirmation")
        else:
            answer = None
            while answer not in ("y", "n"):
                answer = input("Building database at " + arguments['<location>'] + " ? [y/n]")
                if answer == "y":
         # Do this.
                    pass
                elif answer == "n":
         # Do that.
                    pass
                else:
    	            print("Please enter y/n")

            if answer == "n":
                sys.exit(1)

        print( "Building database at " + arguments['<location>'] )
        batchBuild(arguments['<location>'], arguments['--input'], Nsize=int(arguments['--size']))

    if arguments['--info']:
        data = stat(arguments['<location>'])
        ff = None if arguments['--out'] == 'stdout' else open(arguments['--out'], "a")
        print(data, file=ff)

    if arguments['--get']:
        data = get(arguments['--get'], arguments['<location>'])
        ff = None if arguments['--out'] == 'stdout' else open(arguments['--out'], "a")
        print(data, file=ff)
    
    if arguments['--view']:
        preview(arguments['--view'])
    #arguments['location'])

    if arguments['--cluster']:
        print("Cluster mode")
        maxClusterIndex = arguments['--radius'] if arguments['--radius'] else 100000
        setNodes(arguments['<location>'], arguments['--cluster'], maxClusterIndex)
    
    if arguments['--nodes']:

        print (arguments)
        nWorker = arguments['--threads'] if arguments['--threads'] else 8
       
        
        opt = {}
        if arguments['--size']:
            opt['Nsize'] = int(arguments['--size'])
        
        reNumNode = arguments['--nodes'].replace('?', '.').replace('*', '.*')
        reNumNode = '_' + reNumNode + '\.gz$'
        #print(reNumNode)
        inputClusterIter =  [ (iFile, opt) for iFile in glob.iglob(arguments['<location>'] + '/node_*.gz' ) if re.search(reNumNode, iFile) ]
        print(inputClusterIter)
        with Pool(processes=nWorker) as pool:
            pool.map(pClusterBuild, inputClusterIter)
        
        print ("Done")
        #pClusterBuild(arguments['<location>'], number)

        '''
python ~/work/DVL/python3/pyproteinsext/src/pyproteinsext/database/uniprotFastaFS.py ~/tmp/vTrembl --node '??'

        '''


from pyrediscore import *
from .entryProxy import EntrySchema
import progressbar


entryProxyLoader = EntrySchema().load

# Basic storage no account for alternativ ID
# receive uniprot ID
# hash mainID
@connect
@store
def storeEntry(uniprotObj, *args, **kwargs):
    return f"uniprot:{uniprotObj.id}", uniprotObj.toJSON()

@connect
@storeMany
def storeManyEntries(entryList, *args, **kwargs):
    return { f"uniprot:{e.id}" : e.toJSON() for e in entryList}
    
@connect
@delete
def removeEntries(uniprotIDs, *args, **kwargs):
    return [f"uniprot:{_}" for _ in uniprotIDs]
#batch=50
#    for key in listUniprotKey():

@connect
@get
def getUniProtEntry(uniprotID, *args, raw=False, **kwargs):
    return (f"uniprot:{uniprotID}", entryProxyLoader)

@connect
@mget
def mgetUniProtEntry(uniprotIDs, *args, raw=False, **kwargs):
    _ = [ f"uniprot:{uniprotID}" for uniprotID in uniprotIDs ]
    return (_, entryProxyLoader)

@connect
@listKey
def listUniprotKey(*args, prefix=False, **kwargs):
    return ('uniprot:*', 'uniprot:')


"""Uniprot ressources microservice

Usage:
  pyproteinsext service uniprot redis start [<xmlProteomeFile>] [--rh=<redis_host> --rp=<redis_port>] [--silent] [--port=<portNumber>]
  pyproteinsext service uniprot redis wipe [--rh=<redis_host> --rp=<redis_port>]
  pyproteinsext service uniprot xml start <xmlProteomeFile> [--silent] [--port=<portNumber>]
    
Options:
  -h --help     Show this screen.
  --port=<portNumber>  port for public API [default: 2332]
  --rp=<redis_port>  redis DB TCP port [default: 6379]
  --rh=<redis_host>  redis DB http adress [default: localhost]
  --silent  verbosity
  
"""
import os


from docopt import docopt
from .services import uniprot as uniprotService

arguments = docopt(__doc__)
print(arguments)

if arguments['wipe']:
  uniprotService.cleanup(rh=arguments['--rh'], rp=arguments['--rp'])
  exit(1)

app = uniprotService.startup(arguments['<xmlProteomeFile>'],\
    redis=arguments['redis'],\
    rh=arguments['--rh'], rp=arguments['--rp'])

app.run(debug=False, port=arguments['--port'])



import sys, getopt, numpy
import json
import os
import subprocess
from cStringIO import StringIO
from Bio import SearchIO
import pyproteins.sequence.peptide

def main(argv):

    try:
        opts, args = getopt.getopt(argv,"hi:o:",[])

    except getopt.GetoptError:
        print 'pfamCutter.py -i <inputfile> -o <tag for outputfiles>'
        sys.exit(22)
    for opt, arg in opts:
        if opt == '-h':
            print 'pfamCutter.py -i <inputfile> -o <tag>'
            sys.exit()
        elif opt in ("-i", "--ifile"):
            inputfile = arg
        elif opt == "-o":
            outputTag = arg

    c = Cutter()
    c.cut(fileName=inputFile)
    c.write()


class Cutter():
    def __init__(self, hmmrPath="/Users/guillaumelaunay/work/bin/hmmer-3.1b2-macosx-intel/binaries", hmmProfilePath="/Users/guillaumelaunay/work/data/hmm/profiles"):
        self.path = {'profiles' : hmmProfilePath, 'binaries' : hmmrPath }
        self._index()
    def _index(self):
        self.profiles = [ f.replace(".h3i", "") for f in os.listdir(self.path["profiles"]) if f.endswith(".h3i") ]

    def cut(self, **kwargs):
        self.ruler="Q8NF91"
        if 'ruler' in kwargs:
            self.ruler = kwargs['ruler']

        if self.ruler not in self.profiles:
            raise ValueError(ruler + " is not a registred database of hmm profiles")

        if 'fileName' in kwargs:
            print self.path["profiles"] + '/' + self.ruler + ' ' + kwargs['fileName']
            p = subprocess.Popen([ self.path["binaries"] + '/hmmscan', self.path["profiles"] + '/' + self.ruler, kwargs['fileName'] ], stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
            out, err = p.communicate()
            self._parser(out, err)

    def write(self):
        for i, peptide in enumerate(self):
            with open (peptide.id.replace(" ", "_") + "_" + str(i) + ".fasta", "w") as f:
                f.write(peptide.fasta)

    def _parser(self, stdout, stderr):
            f = StringIO(stdout)
            self.results = SearchIO.read(f, "hmmer3-text")

    def __iter__(self):
        for hits in self.results:
            for i, h in enumerate(hits):
                r = h.query_range
                yield pyproteins.sequence.peptide.Entry(id=self.results.id + " " + hits.id + " " + str(r), seq=str(h.query.seq).replace("-", ""))

if __name__ == "__main__":
    main(sys.argv[1:])


import pyproteins.services.utils

#
# Returns
# ("Fields", .... )
#
#
#


# Canonical annotator data structure

#ECMannotator = {
#    'tag' : "ECM",  #Name of the annotator
#        'positiveAnnotationList' : [ #set of properties that selects an entry
#        { 'name' : 'goWords', # Name of the column in csv
#          'target' : 'GO', # properties/Attribute to get from the tested object
#           'content' : [  # Values to test against got attributes
#            { 'id' : 'GO:0016020', 'txt' : 'membrane'}, # a test value, only id is effectively tested
#        ]
#      },
#        { 'name' : 'keyWords',
#          'target' : 'KW',
#          'content' : [
#                { 'id' : 'KW-0472', 'txt' : 'Membrane'},
#            ],
#        }
#    ],
#    'negativeAnnotationList' : [
#        { 'name' : 'badGoWords',
#          'target' : 'GO',
#          'content' : [
#            { 'id' : 'GO:0031965', 'txt' : 'nuclear membrane' }
#            ]
#        }
#    ]
#}

# Annotator type/sub properties checks
def _checkConstraints(datum):
    if 'tag' not in datum:
        raise ValueError("No \'tag\' values in annotator dictionary input")
    if not 'positiveAnnotationList' in datum or not 'negativeAnnotationList' in datum:
        raise ValueError("No \'positiveAnnotationList\' or \'negativeAnnotationList\' values in annotator dictionary input")

def _checkConstraint(data):
    for d in data:
        if not 'name' in data:
            raise ValueError("No \'name\' value in annotation element")
        if not 'target' in data:
            raise ValueError("No \'target\' value in annotation element")
        if not 'content' in data:
            raise ValueError("No \'content\' value in annotation element")
    for d in data['content']:
        if not 'id' in d:
            raise ValueError("No \'id\' value in annotation element content")



#
#
def stringifyContraintList(data):
    _str = ''
    for terms in data:
        #print '-->' + str(terms) + '<--'
        _str += '\tname : ' + terms['name'] + '\n\ttarget : ' + terms['target'] + '\n'
        _str += '\tcontent:\n' + '\n'.join(['\t\t' + x.id + ', ' + x.txt for x in terms['content'] ])
        _str += '\n'
    return _str

class AnnotationTerm(object):
    def __init__(self, datum):
        if 'id' not in datum:
            raise ValueError("missing id in annotation term specs")
        self.id = datum['id']
        self.txt = datum['txt'] if 'txt' in datum else 'N/A'

    def __str__(self):
        #return term['id'] + '(' + term['txt'].replace('(', '\(').replace(')', '\)') + ')'
        return self.id + '(' + self.txt + ')'
    def __repr__(self):
        return str(self)

class Uniprot(object):
    def __repr__(self):
        _str = 'tag : ' + self.tag + '\n'
        if 'positiveAnnotationList' in self.constraints:
            if self.constraints['positiveAnnotationList']:
                _str += 'positive annotation terms:\n'
                _str += stringifyContraintList(self.constraints['positiveAnnotationList'])
        #print _str
        if 'negativeAnnotationList' in self.constraints:
            if self.constraints['negativeAnnotationList']:
                _str += 'negative annotation terms:\n'
                _str += stringifyContraintList(self.constraints['negativeAnnotationList'])

        return _str

    def __init__(self, tag="default"):
        self.tag = tag
        self.constraints = {}

    def loadConstraints(self, datum):
        _checkConstraints(datum)
        if not self.constraints:
            self.constraints = datum

    def addPositive(self, datum):
        self._add(datum, 'positiveAnnotationList')
    def addNegative(self, datum):
        self._add(datum, 'negativeAnnotationList')

    def _add(self, datum, aType):
        if not isinstance(datum, list):
            datum = [datum]
        toAddList = []
        for d in datum:
            _checkConstraint(d)
            toAdd = {'name' : d['name'],
                     'target' : d['target'],
                     'content' :[]}
            toAdd['content'] = [ AnnotationTerm(x) for x in d['content']]
            toAddList.append(toAdd)

        if not aType in self.constraints:
            self.constraints[aType] = []
        self.constraints[aType] += toAddList

    #returns :
    #  termsNameX  : [[value, ..], ...]    #N element annotated
    #  termsNameY  : [[value, ..], ...]    #N element annotated
    #  status : [True, False ... ]         #N element status
    def annotateAll(self, elementList):
        tType = ['positiveAnnotationList', 'negativeAnnotationList']
        data = {}
        status = []
      
        for e in elementList:
           
            mBool, datum = self.annotate(e)
          
            status.append(mBool)
            for aType in tType:
                if not datum[aType]:
                    break
                for d in datum[aType]:
                    if d['name'] not in data:
                        data[d['name']] = []
                    data[d['name']].append(d['matches'])
     
        return (data, status)
            #annotationHits[annotationType]
            #d = { 'name' : annotation['name'], 'matches' : [] }

#   return the annotations found in the entry
    def annotate(self, elem):
        e = None
        if hasattr(elem, 'GO'):
            e = elem
        elif hasattr(elem, '_uniprotBound'):
            e = elem._uniprotBound
        else:
            raise ValueError('Cant find valid uniprot Object, cant annotate')

        if not self.constraints:
            raise ValueError( 'constraints required to perform annotation test')
        (matchBool, matchContent) = testEntry(e, self.constraints)
        return (matchBool, matchContent)
# Returns a boolean if the uniprot entry matches  specified annotation rules
    def isValid(self, e):
        # unroll dictWords, goWords keyWords badGoWords
        if not self.constraints:
            raise ValueError( 'constraints required to perform annotation test')

        (matchBool, matchContent) = testEntry(e._uniprotBound, self.constraints)
        return matchBool

    def pandify(self):
        pass
        #for annotationType in ['positiveAnnotationList', 'negativeAnnotationList']:
        #    for termsList in matchContent[annotationType]:
        #        tag = annotator['tag'] + '_' + termsList['name']
        #        self[tag] = 'NA' if not termsList['matches'] else (';').join(['(ID:' + d['id'] + ')' + d['txt'] for d in termsList['matches']])


def testEntry(uniProtObj, annotator):
    annotationHits = {
        "positiveAnnotationList" : [],
        "negativeAnnotationList" : []
        }
    blackListedTermFound = False
    wishedListedTermFound = False
    for annotationType in ['positiveAnnotationList', 'negativeAnnotationList']:
        if annotationType in annotator:
            for annotation in annotator[annotationType]:
                _testBoundMethod = None
                if annotation['target'] == 'GO':
                    _testBoundMethod = uniProtObj.hasGO
                elif annotation['target'] == 'KW':
                    _testBoundMethod = uniProtObj.hasKW
                elif annotation['target'] == 'MIM':
                    _testBoundMethod = uniProtObj.hasMIM
                elif annotation['target'] == 'DI':
                    _testBoundMethod = uniProtObj.hasDI
                elif annotation['target'] == 'ORPHA':
                    _testBoundMethod = uniProtObj.hasORPHA
                else:
                    raise ValueError("Unknown annotation to test on uniprot Entry " + str(annotation['target']))

                d = { 'name' : annotation['name'], 'matches' : [] }
                for term in annotation['content']:
                    if _testBoundMethod(term.id):
                        d['matches'].append(term)
                        if annotationType == 'positiveAnnotationList':
                            wishedListedTermFound = True
                        else:
                            blackListedTermFound = True
                annotationHits[annotationType].append(d)

            entryBool = False if blackListedTermFound else wishedListedTermFound
    return (entryBool, annotationHits)


class Matrisome:
    def __init__(self, masterFile=None):
        if not masterFile:
            raise ValueError ("You must provide a masterFile argument to Matrisome constructor")

        tsvBuffer = pyproteins.services.utils.tsvToDictList(fileName=masterFile)
        self.data = tsvBuffer['data']
        self.keymap = tsvBuffer['keymap']
        self.accessors = self._index()

    # index entry using relevant key found in their record
    def _index(self):
        accessors = {'uniprot' : {}, 'gene' : {}} # any other meaningfull accessor
        for d in self.data:
            #print "-->" + d['UniProt_IDs'] + '<--'
            if d['UniProt_IDs']:
                for k in d['UniProt_IDs'].split(':'):
                    if k in accessors['uniprot']:
                        print (k + ' found multiple time')
                    else :
                        accessors['uniprot'][k] = []
                    accessors['uniprot'][k].append(d)
            if d['Gene Symbol']:
                for k in d['Gene Symbol'].split(':'):
                    if k in accessors['gene']:
                        print (k + ' found multiple time')
                    else :
                        accessors['gene'][k] = []
                    accessors['gene'][k].append(d)
        return accessors

    def get(self, uniprotID=None):
        if uniprotID:
            if not uniprotID in self.accessors['uniprot']:
                return []
            return [{'Category' : d['Category'], 'Division' : d['Division']} for d in self.accessors['uniprot'][uniprotID]]
        return []